#version 330 core
#pragma include "base.glsl"

#define M_PI 3.1415926535897932384626433832795

layout (location=0) in vec3 position;
layout (location=1) in vec3 normal;
layout (location=2) in vec3 color;
layout (location=3) in vec2 diffuse_texture_coordinates;

uniform mat4 light_view;
uniform mat4 light_projection;
uniform mat4 projection;
uniform mat4 view;
uniform mat4 model;

out vec3 Color;
out vec3 FragPos;
out vec3 Normal;
out vec2 UV;
out vec4 FragPosLightSpace;

#pragma include "morph.glsl"

void main()
{
	vec4 pos = vec4(position.xyz, 1);
	
	morph_model_by_category(pos);	 //morph in model-space
	vec4 world_pos = model * pos;
	morph_world(world_pos);

    gl_Position = projection * view * world_pos;
    Color = color;
    Normal = vec3(normalize(transpose(inverse(model)) * vec4(normal, 1)));
    FragPos = vec3(world_pos);
    UV = diffuse_texture_coordinates;
    FragPosLightSpace = light_projection * light_view * vec4(FragPos, 1);
}


/* desmos:
f\left(x\right)\ =6+5\operatorname{sech}\left(x-3\right)+\ln \left(x\right)-\frac{e^{\frac{x}{10}}}{5}
f\left(x\right)\ =\ 1-e^{-2x}+0.9\operatorname{sech}\left(x-4\right)-0.005e^x
f\left(x\right)\ =\ h\frac{2}{\left(1.1-h+500e^{-x}\right)}-h\cdot 0.005e^{0.2x}
f\left(t\right)\ =\ h\frac{2e^{-h\cdot 0.04t}}{\left(1.1-h+500e^{-t}\right)}-h\cdot 0.005e^{0.2t}  (h is not the highness, but the strength of the shroom multiplicated with your tolerance)
*/