#version 330 core
out vec4 FragColor;

in vec2 TexCoords;

uniform sampler2D scene;
uniform sampler2D bloomBlur;
uniform sampler2D particles;
uniform bool bloom_active;
uniform bool particle_active;
uniform float exposure;

//uniform sampler2D closestDepthDebug;
//uniform sampler2D currentDepthDebug;

float map(float v, float min1, float max1, float min2, float max2);

void main()
{             
    const float gamma = 2.2;
    vec3 hdrColor = texture(scene, TexCoords).rgb;      
    vec3 bloomColor = texture(bloomBlur, TexCoords).rgb;
	vec4 particleColor = texture(particles, TexCoords);
    if(bloom_active) {
        hdrColor += bloomColor;
	}
	if(particle_active) {
		hdrColor += particleColor.rgb*particleColor.w; // additive blending
	}
    // tone mapping
    vec3 result = vec3(1.0) - exp(-hdrColor * exposure);
    // also gamma correct while we're at it       
    //result = pow(result, vec3(1.0 / gamma));
	
	// if(gl_FragCoord.x < 300 && gl_FragCoord.y < 300) {
		// vec2 uv = vec2(map(gl_FragCoord.x, 0.0, 300.0, 0.0, 1.0), map(gl_FragCoord.y, 0.0, 300.0, 0.0, 1.0));
		// result = texture(particles, uv).rgb;
	// }
	
	// if(gl_FragCoord.x > 300 && gl_FragCoord.x < 600 && gl_FragCoord.y < 300) {
		// vec2 uv = vec2(map(gl_FragCoord.x, 300.0, 600.0, 0.0, 1.0), map(gl_FragCoord.y, 0.0, 300.0, 0.0, 1.0));
		// result = texture(bloomBlur, uv).rgb;
		//result = shadow_map;
	// }
	
	/*
	if(gl_FragCoord.x > 600 && gl_FragCoord.x < 900 && gl_FragCoord.y < 300) {
		vec2 uv = vec2(map(gl_FragCoord.x, 600.0, 900.0, 0.0, 1.0), map(gl_FragCoord.y, 0.0, 300.0, 0.0, 1.0));
		result = texture(closestDepthDebug, uv).rgb;
		//result = shadow_map;
	}
	
	if(gl_FragCoord.x > 900 && gl_FragCoord.x < 1200 && gl_FragCoord.y < 300) {
		vec2 uv = vec2(map(gl_FragCoord.x, 900.0, 1200.0, 0.0, 1.0), map(gl_FragCoord.y, 0.0, 300.0, 0.0, 1.0));
		result = texture(currentDepthDebug, uv).rgb;
		//result = shadow_map;
	}*/
	
    FragColor = vec4(result, 1.0);
}

float map(float v, float min1, float max1, float min2, float max2) {
    return (v-min1)/(max1-min1)*(max2-min2)+min2;
}