void calc_point_lights(inout vec3 point_light_sum, in vec3 object_color, in float specular_strength, in vec3 norm, in vec3 view_dir);
void calc_directional_lights(inout vec3 directional_light_sum, in vec3 object_color, in float specular_strength, in vec3 norm, in vec3 view_dir);
void calc_spot_lights(inout vec3 spot_light_sum, in vec3 object_color, in float specular_strength, in vec3 norm, in vec3 view_dir);
void diffuse_specular(inout vec3 diffuse, inout vec3 specular, in vec3 light_color, in float specular_strength, in vec3 light_dir,
                                in vec3 norm, in vec3 light_intensity, in vec3 view_dir);
void attenuation(inout vec3 intensity, in vec3 light_pos, in vec3 light_color, in float constant_attenuation, in float linear_attenuation, in float quadratic_attenuation);
void compute_highness_effect(inout vec3 raw_color);
void shadow_calculation(out float shadow, vec4 frag_pos_light_space, vec3 light_dir);
float omni_shadow_calculation(vec3 fragPos);
float map(float v, float min1, float max1, float min2, float max2);
vec2 map_v2(vec2 v, float min1, float max1, float min2, float max2);

uniform float emission;

void shading(out vec4 frag_color, out vec4 hdr_map/*, out vec4 closest_depth_map, out vec4 current_depth_map*/) {
    vec3 object_color = Color;
    float ambient_strength = 0.0235f;
    float specular_strength = 0.5f;
    vec3 norm = normalize(Normal);
    vec3 view_dir = normalize(view_pos - FragPos);


    if(texture_active) {
        vec4 tex = texture(diffuse_texture, UV);
        vec3 tex_color = vec3(tex[0], tex[1], tex[2]);
        object_color = object_color * tex_color * tex[3];
    }
	
    vec3 point_light_sum = vec3(0,0,0);
    calc_point_lights(point_light_sum, object_color, specular_strength, norm, view_dir);

    vec3 directional_light_sum = vec3(0,0,0);
    calc_directional_lights(directional_light_sum, object_color, 0, norm, view_dir); //specular in directional light will result in ugly bright light

    vec3 spot_light_sum = vec3(0,0,0);
    calc_spot_lights(spot_light_sum, object_color, specular_strength, norm, view_dir);

    vec3 result = point_light_sum + directional_light_sum + spot_light_sum + object_color * (ambient_strength + emission);

	compute_highness_effect(result);

	float brightness = dot(result.rgb, vec3(0.2126, 0.7152, 0.0722));
    
    hdr_map = vec4(result.rgb * clamp(brightness*brightness*brightness*brightness*brightness*brightness*brightness*brightness, 0.0, 1.0), 1.0);
	
	/*
	if(gl_FragCoord.x < 300 && gl_FragCoord.y < 300) {
		vec2 uv = map_v2(gl_FragCoord.xy, 0.0, 300.0, 0.0, 1.0);
		vec4 tex = texture(shadow_map, uv);
        result = vec3(tex[0], tex[0], tex[0]);
		//result = shadow_map;
	}*/
	/*
	vec3 fragToLight = FragPos - spot_light_position[0];
    float currentDepth = length(fragToLight);
    float shadow = 0.0;
    float bias = 0.0005;
    int samples = 20;
    float viewDistance = length(view_pos - FragPos);
    //float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
	float diskRadius = 0.05;
	float closestDepth = 0;
    for(int i = 0; i < samples; ++i)
    {
        closestDepth = texture(omni_map, fragToLight + gridSamplingDisk[i] * diskRadius).r;
		closestDepth *= far_plane;   // Undo mapping [0;1]		
    }
	closest_depth_map = vec4(vec3(closestDepth / far_plane), 1.0);
	current_depth_map = vec4(vec3(currentDepth / far_plane), 1.0);
	//FragColor = vec4(vec3(closestDepth / far_plane), 1.0);   */
    frag_color = vec4(result, 1.0f);		
}


void calc_point_lights(inout vec3 point_light_sum, in vec3 object_color, in float specular_strength, in vec3 norm, in vec3 view_dir) {
    for(int p=0; p<point_light_count; p++) {
        if(point_light_visible[p]) {
            //direction
            vec3 light_dir = normalize(point_light_position[p] - FragPos);

            vec3 light_intensity = vec3(0,0,0);
            attenuation(light_intensity, point_light_position[p], point_light_color[p], point_light_constant_attenuation[p], point_light_linear_attenuation[p], point_light_quadratic_attenuation[p]);

            vec3 diffuse = vec3(0,0,0); vec3 specular = vec3(0,0,0);

            diffuse_specular(diffuse, specular, point_light_color[p], specular_strength, light_dir, norm, light_intensity, view_dir);

            vec3 result = (diffuse + specular) * object_color * point_light_power[p];
            point_light_sum += result;
        }
    }
}

void calc_directional_lights(inout vec3 directional_light_sum, in vec3 object_color, in float specular_strength, in vec3 norm, in vec3 view_dir) {
    for(int d=0; d<directional_light_count; d++) {
        if(directional_light_visible[d]) {
            //direction
            vec3 light_dir = normalize(directional_light_direction[d]);

            //intensity
            vec3 light_intensity = directional_light_color[d]; //* directional_light_intensity[d] comes at the end

            vec3 diffuse = vec3(0,0,0); vec3 specular = vec3(0,0,0);

            diffuse_specular(diffuse, specular, directional_light_color[d], specular_strength, light_dir, norm, light_intensity, view_dir);

			float directional_shadow = 0;
			shadow_calculation(directional_shadow, FragPosLightSpace, light_dir);
			
            vec3 result = ((1.0-directional_shadow) * (diffuse + specular)) * object_color * directional_light_intensity[d];
            directional_light_sum += result;
        }
    }
}

void calc_spot_lights(inout vec3 spot_light_sum, in vec3 object_color, in float specular_strength, in vec3 norm, in vec3 view_dir) {
	float shadow = omni_shadow_calculation(FragPos);

    for(int s=0; s<spot_light_count; s++) {
        if(spot_light_visible[s]) {
            vec3 light_dir = normalize(spot_light_position[s] - FragPos);
            vec3 sd = normalize(vec3(spot_light_direction[s]));
	    float fCosine = dot(sd,light_dir);
	    float fDif = 1.0-spot_light_cutoff[s]; 
  	    float fFactor = clamp((fCosine-spot_light_cutoff[s])/fDif, 0.0, 1.0);
            // inside the cone?
            if (fCosine > spot_light_cutoff[s]) {
                vec3 light_intensity = vec3(0,0,0);
                attenuation(light_intensity, spot_light_position[s], spot_light_color[s], spot_light_constant_attenuation[s], spot_light_linear_attenuation[s], spot_light_quadratic_attenuation[s]);

                vec3 diffuse = vec3(0,0,0); vec3 specular = vec3(0,0,0);

                diffuse_specular(diffuse, specular, spot_light_color[s], specular_strength, light_dir, norm, light_intensity, view_dir);

                vec3 result = ((1.0-shadow) * (diffuse + specular)) * object_color * spot_light_intensity[s] * fFactor;
                spot_light_sum += result;
            }
        }
		shadow = 0.0;
    }
}

void diffuse_specular(inout vec3 diffuse, inout vec3 specular,
                                in vec3 light_color, in float specular_strength, in vec3 light_dir,
                                in vec3 norm, in vec3 light_intensity, in vec3 view_dir)
{
    //diffuse part
    float l_dot_n = clamp(dot(light_dir, norm), 0.0f, 1.0f);
    diffuse = light_intensity * l_dot_n; //light_intensity includes the light_color

    //specular part
    vec3 reflect_dir = reflect(-light_dir, norm);
    float spec = pow(max(dot(view_dir, reflect_dir), 0.0), 2.f); //32 is the shininess parameter
    specular = specular_strength * spec * light_color;
}

void attenuation(inout vec3 intensity, in vec3 light_pos, in vec3 light_color, in float constant_attenuation, in float linear_attenuation, in float quadratic_attenuation) {
    float d = length(FragPos - light_pos);
    float attenuation = constant_attenuation + d*linear_attenuation + d*d*quadratic_attenuation;
    intensity = light_color / attenuation;
}

void compute_highness_effect(inout vec3 raw_color) {

	raw_color.x = raw_color.x * (1 - highness * pow(sin(2.9f * highness/2.f ), 2.f) * pow(sin(0.05f * FragPos.x),2.f) * pow(sin(0.02f * FragPos.y + 1),2.f) * pow(sin(0.008f * FragPos.z + 1.3),2.f));
	raw_color.y = raw_color.y * (1 - highness * pow(sin(2.9f * highness/2.f + 1.f), 2.f));
	raw_color.z = raw_color.z * (1 - highness * pow(sin(2.3f * highness/2.f + 4.33f), 2.f) * pow(sin(0.05f * FragPos.y + 0.7),2.f));
}

void shadow_calculation(out float shadow, vec4 frag_pos_light_space, vec3 light_dir) {
    // perform perspective divide
    vec3 projCoords = frag_pos_light_space.xyz / frag_pos_light_space.w;
    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;
	
    if(projCoords.z > 1.0) {
        shadow = 0.0;
		return;
	}
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    //float closestDepth = texture(shadow_map, projCoords.xy).r; 
	
	float bias = 0.0003 * tan(acos(dot(Normal, light_dir))); // cosTheta is dot( n,l ), clamped between 0 and 1
	bias = clamp(bias, 0, 0.0005);
	float currentDepth = projCoords.z;
	shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadow_map, 0);
	
	//PCF!!!
	for(int x = -1; x <= 1; ++x)
	{
		for(int y = -1; y <= 1; ++y)
		{
			float pcfDepth = texture(shadow_map, projCoords.xy + vec2(x, y) * texelSize).r; 
			shadow += currentDepth - bias > pcfDepth ? 1.0 : 0.0;   
			//shadow += currentDepth > pcfDepth  ? 1.0 : 0.0;			
		}    
		shadow /= 9.0;
	}
    // get depth of current fragment from light's perspective
   
    // check whether current frag pos is in shadow
	//float bias = max(0.05 * (1.0 - dot(Normal, light_dir)), 0.005);
	
	
	//shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;  
	//shadow = currentDepth > closestDepth  ? 1.0 : 0.0;
}  

float omni_shadow_calculation(vec3 fragPos)
{
    vec3 fragToLight = fragPos - spot_light_position[0];
    float currentDepth = length(fragToLight);
    float shadow = 0.0;
    float bias = 0.03;
    int samples = 20;
    float viewDistance = length(view_pos - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(omni_map, fragToLight + gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= far_plane;   // undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadow += 1.0;
    }
    shadow /= float(samples);
        
    // display closestDepth as debug (to visualize depth cubemap)
    // FragColor = vec4(vec3(closestDepth / far_plane), 1.0);    
        
    return shadow;
}

float map(float v, float min1, float max1, float min2, float max2) {
    return (v-min1)/(max1-min1)*(max2-min2)+min2;
}

vec2 map_v2(vec2 v, float min1, float max1, float min2, float max2) {
    return vec2(map(v.x, min1, max1, min2, max2), map(v.y, min1, max1, min2, max2));
}
